"""
CARGO

A data-centric transformation of monoliths into microservices
"""

__version__ = "0.1.0"
__author__ = 'Rahul Krishna <imralk+oss@gmail.com> and Vikram Nitin <vikramnitin9@gmail.com>'
__credits__ = 'IBM Research :: Columbia University'

from .cargo import Cargo
__all__ = ['Cargo']
